import java.awt.BorderLayout;
import java.awt.Button;
import java.awt.Checkbox;
import java.awt.Choice;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.FileDialog;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GraphicsEnvironment;
import java.awt.Label;
import java.awt.Panel;
import java.awt.RenderingHints;
import java.awt.SystemColor;
import java.awt.TextField;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;
import java.awt.image.MemoryImageSource;
import java.awt.image.Raster;
import java.awt.image.RenderedImage;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.ByteOrder;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.imageio.IIOImage;
import javax.imageio.ImageTypeSpecifier;
import javax.imageio.ImageWriteParam;
import javax.imageio.ImageWriter;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.spi.ImageWriterSpi;
import javax.imageio.stream.FileImageOutputStream;
import javax.imageio.stream.ImageOutputStream;

/**
 * A primitive tga writer class 
 * @author Eike Decker
 *
 */
class TGAWriter extends ImageWriter {
	protected TGAWriter(ImageWriterSpi originatingProvider) {
		super(originatingProvider);
	}

	@Override
	public IIOMetadata getDefaultStreamMetadata(ImageWriteParam param) {
		System.out.println("Called 1?");
		return null;
	}

	@Override
	public IIOMetadata getDefaultImageMetadata(ImageTypeSpecifier imageType, ImageWriteParam param) {
		System.out.println("Called 2?");
		return null;
	}

	@Override
	public IIOMetadata convertStreamMetadata(IIOMetadata inData, ImageWriteParam param) {
		System.out.println("Called 3?");
		return null;
	}

	@Override
	public IIOMetadata convertImageMetadata(IIOMetadata inData, ImageTypeSpecifier imageType, ImageWriteParam param) {
		System.out.println("Called 4?");
		return null;
	}

	@Override
	public void write(IIOMetadata streamMetadata, IIOImage image, ImageWriteParam param) throws IOException {
		ImageOutputStream out = (ImageOutputStream) getOutput();
		RenderedImage ri = image.getRenderedImage();
		byte bytes[] = {
			0,0,10,//2,// idlength, colourmaptype,datatype = compressed
			0,0, 
			0,0, 0, 0,0, 0,0, // header until width
		};
		out.reset();
		out.setByteOrder(ByteOrder.LITTLE_ENDIAN);
		out.write(bytes);
		out.writeShort(ri.getWidth());
		out.writeShort(ri.getHeight());
		out.writeByte(32);
		out.writeByte(0);
		
		Raster r = ri.getData();

		int [] data = r.getPixels(0,0,ri.getWidth(),ri.getHeight(),(int[])null);
		byte output[] = new byte[data.length];
		final int w = ri.getWidth(),h = ri.getHeight();
		for (int x=0;x<w;x++) {
			for (int y=0;y<h;y++) {
				//System.out.println(y);
				int a = x*4 + (h-y-1)*4*w, b = x*4+y*w*4;
				output[a+2] = (byte) data[b];
				output[a+1] = (byte) data[b+1];
				output[a+0] = (byte) data[b+2];
				output[a+3] = (byte) data[b+3];
			}
		}
		byte compressed[] = new byte[data.length*2];
		int compressedpos = 0;
		for (int i=0;i<output.length/4;i++) {
			int prefetch = i;
			while (prefetch+4<output.length/4 && prefetch-i<126&& 
					output[prefetch*4+4] == output[i*4+0] &&
					output[prefetch*4+5] == output[i*4+1] &&
					output[prefetch*4+6] == output[i*4+2] && 
					output[prefetch*4+7] == output[i*4+3]) prefetch ++;
			if (prefetch == i) { // raw!
				prefetch = i+1;
				while (prefetch+4<output.length/4 && prefetch-i<126 &&
						(output[prefetch*4+4] != output[prefetch*4+0] ||
						output[prefetch*4+5] != output[prefetch*4+1] ||
						output[prefetch*4+6] != output[prefetch*4+2] || 
						output[prefetch*4+7] != output[prefetch*4+3])) prefetch++;
				
				compressed[compressedpos++] = (byte)(prefetch-i-1);
				while (i<prefetch) {
					compressed[compressedpos++] = output[i*4+0];
					compressed[compressedpos++] = output[i*4+1];
					compressed[compressedpos++] = output[i*4+2];
					compressed[compressedpos++] = output[i*4+3];
					i++;
				}
				i--;
			} else {
				compressed[compressedpos++] = (byte)(prefetch - i + 127 + 1);
				compressed[compressedpos++] = output[i*4+0];
				compressed[compressedpos++] = output[i*4+1];
				compressed[compressedpos++] = output[i*4+2];
				compressed[compressedpos++] = output[i*4+3];
				i = prefetch;
			}
		}
		out.write(compressed,0,compressedpos);
		
	}
	
}

public class FontGenerator extends Frame{
	private static final long serialVersionUID = 1869257347645249593L;
	

	private Component addTo(Container a, Component b, int width, int height) {
		b.setPreferredSize(new Dimension(width,height));
		a.add(b);
		return b;
	}
	int sqsize = 256,fonttablesizevalue = 16;
	int fontlineheight = 0;
	Choice bitmapsize,fonttablesize;
	Choice font;
	TextField fontsize,offsetX,offsetY;
	Checkbox bold,italic,antialiased;
	BufferedImage fontscreen = new BufferedImage(sqsize,sqsize,BufferedImage.TYPE_INT_ARGB);
	int widths[];
	int chartable[] = new int[256];
	ActionListener tableupdate;
	
	public void updatefontscreen () {
		updatefontscreen(Color.BLACK);
	}
	public void updatefontscreen( Color col ) {
		fontscreen = new BufferedImage(sqsize,sqsize,BufferedImage.TYPE_INT_ARGB);
		fontpanel.setPreferredSize(new Dimension(sqsize,sqsize));
		widths = new int[256];
		
		Graphics2D g = (Graphics2D) fontscreen.getGraphics();
		if (antialiased.getState())
			g.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
		String fname = font.getSelectedItem();
		
		Font f = new Font(fname,(bold.getState()?Font.BOLD:0)|(italic.getState()?Font.ITALIC:0),Integer.parseInt(fontsize.getText()));
		
		g.setFont(f);
		int cid = 0;
		g.setColor(col);
		
		fontlineheight = g.getFontMetrics().getHeight();
		
		int offx = Integer.parseInt(offsetX.getText()),offy = Integer.parseInt(offsetY.getText());
		int idx = 0;
		int inc = sqsize/fonttablesizevalue;
		for (int y = 0;y<fonttablesizevalue; y++) {
			for (int x = 0;x<fonttablesizevalue; x++) {
				char c = (char) chartable[cid++];
				widths[idx++] = g.getFontMetrics().charWidth(c);
				g.drawString(""+c,x*inc+offx,y*inc+g.getFontMetrics().getAscent()+offy);
			}
		}
		
		
		
		g.dispose();
	}
	
	public void load (String path, String file) {
		if (path == null) return;
		File f = new File(path,file);
		try {
			BufferedReader br = new BufferedReader(new FileReader(f));
			String line;
			Pattern p = Pattern.compile("local (\\w+)\\s*=\\s*(.*)");
			for (line = br.readLine();line!=null;line = br.readLine()) {
				Matcher m = p.matcher(line);
				if (m.matches()) {
					String name = m.group(1);
					String value = m.group(2).trim();
					if (name.equals("version") && !value.equals("1.0")) throw new IOException("invalid version");
					if (name.equals("font")) this.font.select(value.substring(1,value.length()-1));
					if (name.equals("size")) this.fontsize.setText(value);
					if (name.equals("antialiased")) this.antialiased.setState(value.equals("true"));
					if (name.equals("italic")) this.italic.setState(value.equals("true"));
					if (name.equals("bold")) this.bold.setState(value.equals("true"));
					if (name.equals("xoffset")) this.offsetX.setText(value);
					if (name.equals("yoffset")) this.offsetY.setText(value);
					if (name.equals("resolution")) this.sqsize = Integer.parseInt(value);
					if (name.equals("tablesize")) {
						this.fonttablesizevalue = Integer.parseInt(value);
						fonttablesize.select(value+"x"+value);
					}
					if (name.equals("chartable")) {
						Pattern p2 = Pattern.compile("(\\d+),\\s*");
						int idx = 0;
						for (line = br.readLine();line!=null;line = br.readLine()) {
							
							m = p2.matcher(line);
							while (m.find()) {
								chartable[idx++] = Integer.parseInt(m.group(1));
							}
							if (line.contains("}")) break;
						}
						tableupdate.actionPerformed(null);
					}
				}
			}
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
	}
	
	public void save (String path,String file) {
		if (path == null) return;
		File f = new File(path,file);
		try {
			BufferedWriter bw = new BufferedWriter(new FileWriter(f));
	
			bw.write("-----------------------------------------\n");
			bw.write("-- autogenerated font description file --\n");
			bw.write("-----------------------------------------\n");
			bw.write("local version = 1.0\n");
			bw.write("local font = '"+font.getSelectedItem()+"'\n");
			bw.write("local antialiased = "+(antialiased.getState()?"true":"false")+"\n");
			bw.write("local resolution = "+sqsize+"\n");
			bw.write("local tablesize = "+fonttablesizevalue+"\n");
			bw.write("local italic = "+(italic.getState()?"true":"false")+"\n");
			bw.write("local bold = "+(bold.getState()?"true":"false")+"\n");
			bw.write("local size = "+fontsize.getText()+"\n");
			bw.write("local xoffset = " + offsetX.getText()+"\n");
			bw.write("local yoffset = " + offsetY.getText()+"\n");
			bw.write("local chartable = {\n\t");
			for (int i = 0; i < chartable.length; i++) {
				bw.write(chartable[i]+", ");
				if (i%16==15) bw.write("\n\t");
			}
			bw.write("}\n");
			
			bw.write("return { version = version, " +
						"chartable = chartable, " +
						"name = font, " +
						"xoffset = xoffset, yoffset = yoffset, " +
						"bold = bold, italic = italic, lineheight = "+fontlineheight+", "+
						"size = size,\n");
			bw.write("\ttablesize = tablesize, resolution = resolution,\n");
			bw.write("\twidths = {\n\t");
			for (int i = 0; i < widths.length; i++) {
				bw.write(""+widths[i]+", "+(i%16==15?"\n\t":""));
			}
			bw.write("\n}}");
			bw.close();
			
			file = file.replaceAll("\\..*$",".tga");
			
			ImageWriter iw = new TGAWriter(null);
			FileImageOutputStream out = new FileImageOutputStream(new File(path,file)); 
			iw.setOutput(out);
			updatefontscreen( Color.WHITE);
			
			iw.write(fontscreen);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	@SuppressWarnings("serial") 
	Panel fontpanel = new Panel () {
		{
			setPreferredSize(new Dimension(256,256));
		}
		public void paint(Graphics g) {
			super.paint(g);
			updatefontscreen();
			g.drawImage(fontscreen,0,0,this);
			g.setColor(Color.RED);
			int inc = sqsize / fonttablesizevalue;
			for (int x = 0; x <= fonttablesizevalue; x++) {
				g.drawLine(x*inc,0,x*inc,sqsize);
				g.drawLine(0,x*inc,sqsize,x*inc);
			}
		}
	};
	
	private static int icon[];
	static { // icon initialisation, just read a small icon in hex and replaced the numbers...
		int a = 0xff000000,	b = 0xff010101,	c = 0xff020202,	d = 0xff030303,
			e = 0xff040404,	f = 0xff050505,	g = 0xff060606,	h = 0xff070707,
			i = 0xff080808,	j = 0xff090909,	k = 0xffa0a0a0,	l = 0xffb0b0b0,
			m = 0xffc0c0c0,	n = 0xffd0d0d0,	o = 0xffe0e0e0,	p = 0xffffffff;
		icon = new int[] {
			p,p,p,p,p,o,p,p,p,p,p,p,p,p,p,p,
			p,a,a,a,c,j,k,l,l,j,f,a,a,a,a,p, 
			p,a,a,i,n,p,p,p,p,p,p,k,b,a,a,p, 
			p,a,j,o,p,p,p,p,p,p,p,p,l,b,a,p,
			p,g,p,p,p,p,p,l,i,d,a,c,i,h,a,p,
			p,l,o,p,p,n,i,a,a,a,a,a,a,a,a,p,
			p,p,p,p,p,g,a,a,a,a,a,a,a,a,a,p,
			p,p,p,p,j,a,a,a,a,a,a,a,a,a,a,p,
			p,p,p,o,d,a,a,a,a,h,a,a,h,b,a,p,
			p,o,p,l,a,a,a,a,a,b,a,j,o,k,f,p,
			p,n,o,k,a,a,a,a,a,a,a,h,p,k,a,p,
			p,j,p,k,a,a,d,g,c,a,a,g,e,i,a,p,
			p,a,l,m,a,a,h,p,j,a,a,a,a,a,a,p,
			p,a,c,l,d,a,e,j,a,a,a,a,a,a,a,p,
			p,a,a,a,f,a,a,a,a,a,a,a,a,a,a,p,
			p,p,p,p,p,p,p,p,p,p,p,p,p,p,p,p,
		};
		int in [] = new int[16*16]; // oops, inverted from hex import...
		for (int x=0;x<16;x++)
			for (int y=0;y<16;y++)
				in[x+16*y] = icon[16*15+x-y*16 ];
		icon = in;
	}
	
	public FontGenerator () {
		super("Luxinia Font Bitmap Generator 1.0");

		setIconImage(createImage(new MemoryImageSource(16,16,icon,0,16)));
		addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
				System.exit(0);
			}
		});
		
		setResizable(false);
		
		
		setLayout(new BorderLayout());
		
		Panel west = new Panel();
		
		west.setPreferredSize(new Dimension(200,256));
		add(west,BorderLayout.WEST);
		addTo(west,new Label("Resolution:",Label.RIGHT),80,20);
		bitmapsize = (Choice)addTo(west,new Choice(),116,20);
		bitmapsize.setPreferredSize(new Dimension(98,20));
		bitmapsize.add("64x64");
		bitmapsize.add("128x128");
		bitmapsize.add("256x256");
		bitmapsize.add("512x512");
		bitmapsize.add("1024x1024");
		bitmapsize.select("256x256");
		bitmapsize.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				String item = bitmapsize.getSelectedItem().replaceAll("x.*","");
				sqsize = Integer.parseInt(item);
				updatefontscreen();
				pack();
			}});
		addTo(west,new Label("Tablesize:",Label.RIGHT),80,20);
		fonttablesize = (Choice)addTo(west,new Choice(),98,20);
		fonttablesize.add("6x6");
		fonttablesize.add("8x8");
		fonttablesize.add("16x16");
		fonttablesize.select("16x16");
		fonttablesize.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				String item = fonttablesize.getSelectedItem().replaceAll("x.*","");
				fonttablesizevalue = Integer.parseInt(item);
				updatefontscreen();
				fontpanel.repaint();
			}
		});
		
		font = (Choice)addTo(west,new Choice(),180,20);
		{
			String fontnames[] = GraphicsEnvironment.getLocalGraphicsEnvironment().getAvailableFontFamilyNames();
			
			for (int i = 0; i < fontnames.length; i++) {
				font.add(fontnames[i]);
				//font.add(fonts[i].getName());
			}
			font.select("Arial");
		}
		
		addTo(west,new Label("Font Size:",Label.RIGHT),114,20);
		fontsize = (TextField)addTo(west,new TextField(),60,20);
		fontsize.setText("10");
		ActionListener actionrep = new ActionListener(){public void actionPerformed(ActionEvent e) {fontpanel.repaint();}};
		fontsize.addActionListener(actionrep);
		
		bold = (Checkbox)addTo(west,new Checkbox("Bold"),180,20);
		italic = (Checkbox)addTo(west,new Checkbox("Italic"),180,20);
		antialiased= (Checkbox)addTo(west,new Checkbox("Antialias"),180,20);
		
		add(fontpanel,BorderLayout.CENTER);
		
		//setPreferredSize(new Dimension(400,280));
		

		ItemListener il = new ItemListener () {
			public void itemStateChanged(ItemEvent e) {
				fontpanel.repaint();
			}			
		};
		font.addItemListener(il);
		bold.addItemListener(il);
		italic.addItemListener(il);
		antialiased.addItemListener(il);
		
		addTo(west,new Label("X Offset:",Label.RIGHT),160,20);
		offsetX = (TextField) addTo(west,new TextField(),20,20);
		offsetX.setText("2");
		addTo(west,new Label("Y Offset:",Label.RIGHT),160,20);
		offsetY = (TextField) addTo(west,new TextField(),20,20);
		offsetY.setText("2");
		offsetX.addActionListener(actionrep);
		offsetY.addActionListener(actionrep);
		
		Button about = (Button) addTo(west,new Button("About"),60,24);
		about.addActionListener(new ActionListener() {
			int h = 0;
			void add(Dialog d,String line) {
				Label ta = new Label(line,Label.CENTER);
				ta.setPreferredSize(new Dimension(300,16));
				h+=20;
				d.add(ta);
			}
			public void actionPerformed(ActionEvent e) {
				final Dialog d = new Dialog(FontGenerator.this,"About",true);
				
				d.addWindowListener(new WindowAdapter() {
					public void windowClosing(WindowEvent arg0) {
						d.dispose();
					};	
				});
				
				d.setLayout(new FlowLayout());
				
				add(d,"Font Bitmap Generator for Luxinia");
				add(d,"Version 1.0");
				add(d,"http://www.luxinia.de");
				add(d,"Written by Eike Decker");
				
				Button ok = new Button("OK");
				ok.setPreferredSize(new Dimension(100,24));
				d.add(ok);
				ok.addActionListener(new ActionListener() {
					public void actionPerformed(ActionEvent arg0) {
						d.dispose();
					}
				});
				d.setPreferredSize(new Dimension(300,h+30+40));
				d.setResizable(false);
				d.pack();
				
				int x = FontGenerator.this.getX();
				int y = FontGenerator.this.getY();
				int w = FontGenerator.this.getWidth();
				int h = FontGenerator.this.getHeight();
				int w1 = d.getWidth();
				int h1 = d.getHeight();
				d.setBounds(x+(w-w1)/2, y + (h-h1)/2, w1, h1);
				d.setVisible(true);
			}
		});
		
		Button load = (Button) addTo(west,new Button("Load"),60,24);
		load.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				FileDialog fd = new FileDialog(FontGenerator.this,"Chose lua font file",FileDialog.LOAD);
				
				fd.setVisible(true);
				String file = fd.getFile();
				FontGenerator.this.load(fd.getDirectory(),file);
			}
		});
		Button save = (Button) addTo(west,new Button("Save"),60,24);
		west.setBackground(SystemColor.menu);
		save.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				FileDialog fd = new FileDialog(FontGenerator.this,"Chose location",FileDialog.SAVE);
				
				fd.setVisible(true);
				String file = fd.getFile();
				if (file == null) return;
				if (!file.matches("^.*\\.lua$"))
					file = file.replaceAll("\\.[^\\.]*$","") + ".lua";
				FontGenerator.this.save(fd.getDirectory(),file);
			}});
		
		
		Panel east = new Panel(new FlowLayout(FlowLayout.CENTER,2,2));
		east.setBackground(SystemColor.menu);
		addTo(east,new Label("Charactertable",Label.CENTER),120,16);
		addTo(east,new Label("# | char # | char",Label.CENTER),120,16);
		
		final Label nums[] = new Label[8];
		final TextField codes[] = new TextField[nums.length];
		final TextField chars[] = new TextField[nums.length];
		final int pos[] = {0};
		final Button prev = new Button("< prev");
		final Button next = new Button("next >");
		final ActionListener upd;
		
		upd = new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				int o = pos[0];
				prev.setEnabled(o>0);
				next.setEnabled(o<fonttablesizevalue*fonttablesizevalue-nums.length);
				for (int i = 0; i < nums.length; i++) {
					if (o+i<fonttablesizevalue*fonttablesizevalue) {
						nums[i].setText(o+i+":");
						codes[i].setText(""+chartable[i+o]);
						chars[i].setText(""+(char)chartable[i+o]);
						codes[i].setEnabled(true);
						chars[i].setEnabled(true);
					} else {
						nums[i].setText("");
						codes[i].setText("");
						chars[i].setText("");
						codes[i].setEnabled(false);
						chars[i].setEnabled(false);
					}
					
				}
			}
		};
		tableupdate = upd;
			
		for (int i=0;i<nums.length;i++) {
			final int j = i;
			final ActionListener tlist = new ActionListener(){
				public void actionPerformed(ActionEvent e)  
				{			
					if (e.getSource() == codes[j]) {
						try {
							chartable[j + pos[0]] = Integer.parseInt(codes[j].getText());
							chars[j].setText(""+(char)chartable[j+pos[0]]);
						} catch (Exception x) {}
						fontpanel.repaint();
						if (j<nums.length-1) {
							codes[j+1].requestFocus();
						} else {
							pos[0]+=nums.length;
							chars[0].requestFocus();
							upd.actionPerformed(null);
						}
					}
					else
					if (chars[j].getText().length()>0){
						chartable[j+pos[0]] = chars[j].getText().charAt(0);
						codes[j].setText(""+chartable[j+pos[0]]);
						fontpanel.repaint();
						if (j<nums.length-1) {
							chars[j+1].requestFocus();
						} else {
							pos[0]+=nums.length;
							chars[0].requestFocus();
							upd.actionPerformed(null);
						}
					}
				}
			};
			FocusListener fl = new FocusAdapter () {
				public void focusLost(FocusEvent e) {
					//tlist.actionPerformed(new ActionEvent(e.getSource(),e.getID(),"none of your business"));
				}
				public void focusGained(java.awt.event.FocusEvent e) {
					((TextField)e.getSource()).selectAll();
				};	
			};
			
			nums[i] = (Label) addTo(east,new Label(i+":",Label.RIGHT),30,22);
			codes[i] = (TextField) addTo(east,new TextField(),30,22);
			chars[i] = (TextField) addTo(east,new TextField(),30,22);
			codes[i].addActionListener(tlist);
			chars[i].addActionListener(tlist);
			codes[i].addFocusListener(fl);
			chars[i].addFocusListener(fl);
		}
		for (int i = 0; i < chartable.length; i++) {
			chartable[i] = i;
		}
		addTo(east,prev,40,22);
		addTo(east,next,40,22);
		
		
		upd.actionPerformed(null);
		
		next.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				pos[0]+=nums.length;
				upd.actionPerformed(null);
			}			
		});
		
		prev.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				pos[0]-=nums.length;
				upd.actionPerformed(null);
			}			
		});
		east.setPreferredSize(new Dimension(100,20));
		add(east,BorderLayout.EAST);
		
		pack();
		setMinimumSize(getSize());
		
		setVisible(true);
	}
	
	public static void main(String[] args) {
		new FontGenerator();
	}
}
