-- Collision COL LUA Exporter
--
-- by Christoph 'CrazyButcher' Kubisch 
-- http://crazybutcher.cube3d.de

-- to be placed in /scripts/startup

-- Custom Attributes
--  
global colnode = attributes CollisionNode attribID:#(0x78433dea, 0x2015d87)
(
	parameters params rollout:mainRoll 
	(
		mass type:#float ui:mass default:0 animatable:false
	)
	rollout mainRoll "Collision Body"
	(
		group	"Collision Properties"
		(
			spinner mass "Mass: "
		)
		-- todo ik limits
	)
)


-- Export Utility
--  

utility ColExport "Collision LUA Export" 
(
	global colExportSelected

	local col
	local colSurfaces = #()
	local outfile
	

	
	
	function colExportSelected outfileName prescale=
	(
		local xmlheader = "--3dsmax generated collision .lua file\n"
		local versionstring = "\"luxinia_collision_data version 001\""
		
		struct InstanceOf (self,parent)
		
		local areinstances = #() 
		
		function stringindent indent =
		(
			local string = ""
			for i=1 to indent do string = string + "\t"
			return string
		)
		
		-- neilblevins soulburn scripts
		fn sLibGetInstancesOfObject Obj = 
		(
			a = #()
			if (refs.dependents Obj.baseobject).count != 0 then
				(
				for i = 1 to (refs.dependents Obj.baseobject).count do
					(
					if (refs.dependents Obj.baseobject)[i] != Obj then
						(
						if iskindof (refs.dependents Obj.baseobject)[i] (classof Obj) then
							(
							append a (refs.dependents Obj.baseobject)[i]
							)
						)
					)
				)
			return a
		)
		
		function isinstanced instancenodes Obj =
		(
			for n in instancenodes do (
				if (n.self == Obj) then (
					return n.parent	
				)
			)
			return undefined
		)
		
		function isparentinstance instancenodes Obj =
		(
			for n in instancenodes do (
				if (n.parent == Obj) then (
					return n.self	
				)
			)
			return undefined
		)
	
		
		if (outfileName != undefined) then
			outfile = createfile outfileName
		
		if (outfile != undefined) then
		(
		format "%" xmlheader to:outfile
		format "%" ("return "+versionstring+",{\n") to:outfile
		
		
		for obj in selection do (
			local class = classOf obj.baseObject
			if (class ==Editable_Poly or class == Editable_mesh) then(
				local list = (sLibGetInstancesOfObject obj)
				
				for child in list do (
					if (undefined == isparentinstance areinstances child) then(
						append areinstances (InstanceOf self:child parent:obj)
					)
				)
			)	
		)
		
		for obj in selection do	(
			
			local class = classOf obj.baseObject
			local tagname 
			local parentscale
			
			case (class) of
			(
				Box:			tagname = "dgeombox"
				Sphere:			tagname = "dgeomsphere"
				Capsule:		tagname = "dgeomccylinder"
				Cylinder:		tagname = "dgeomcylinder"
				Editable_mesh:	tagname	= "dgeomtrimesh"
				Editable_Poly:	tagname = "dgeomtrimesh"
			)
			
			if (tagname == undefined) then continue
			
			-- the usual stuff type and name
			format "%" ((stringindent 1) + "{ type = \""+tagname+"\","+ "name = \"" + obj.name+"\",") to:outfile
			if (obj.parent != undefined) then(
				format "%" ("parent = \""+obj.parent.name+"\",") to:outfile
				parentscale = [1,1,1]
				local nextparent = obj.parent
				while (nextparent) do
				(
					parentscale *= nextparent.scale
					nextparent = nextparent.parent	
				)
			)
			else (
				parentscale = [1,1,1]
			)
			
			-- position and rotation
			-- transforms
			local size
			local position = (in coordsys parent obj.pos) * parentscale * prescale
			local rotation = (in coordsys parent obj.rotation)
			local objscale = in coordsys parent obj.scale
			
			local roteuler = quatToEuler rotation order:6
			
			--roteuler.x *= -1
			size = roteuler.x
			roteuler.x = -roteuler.z
			roteuler.z = -size
			roteuler.y *= -1
			--roteuler.z *= -1

			objscale *= prescale
			objscale *= parentscale
			
			case (class) of
			(
				Box:
				(
					position = (in coordsys parent obj.center) * parentscale * prescale
				)
				Cylinder:
				(
					position = (in coordsys parent obj.center) * parentscale * prescale
				)
				Capsule:
				(
					position = (in coordsys parent obj.center) * parentscale * prescale
				)
			)
			
			local positionstring = "{" + (position.x as string) + "," + (position.y as string) + "," + (position.z as string) + "}"
			local rotstring = "{" + (roteuler.x as string) + "," + (roteuler.y as string) + "," + (roteuler.z as string) + "}"
			
			format "%"("pos = "+positionstring+",") to:outfile
			format "%"("rot = "+rotstring+",\n") to:outfile
			
			-- parameters
			format "%" ((stringindent (2)) + "parameters = { ") to:outfile
			case (class) of
			(
				Box:
				(	
					tagname = "{"+ ((obj.width*objscale.x) as string) + "," + ((obj.length*objscale.y) as string) + "," + ((obj.height*objscale.z) as string) + "}"
					format "%" ("sizes = "+tagname) to:outfile
				)
				Sphere:
				(
					format "%" ("radius = "+((obj.radius*objscale.x) as string)) to:outfile
				)
				Cylinder:
				(
					format "%" ("radius = "+((obj.radius*objscale.x) as string)+",") to:outfile
					
					format "%" ("height = "+((obj.height*objscale.z) as string)) to:outfile
				)
				Capsule:		
				(
					format "%" ("radius = "+((obj.radius*objscale.x) as string)+",") to:outfile
					if (obj.heighttype > 0) then(
						size = obj.height
					)
					else(
						size = obj.height-2*obj.radius
					)
					format "%" ("height = "+((size*objscale.z) as string)) to:outfile
				)
			
			)
			
			if (class == Editable_mesh or class == Editable_Poly) then
			(
				local parent = isinstanced areinstances obj
				if ( parent != undefined) then(
					format "%" ("trimesh = \""+parent.name+"\"") to:outfile
				)
				else (
					local trimesh = obj.mesh
					format "%" ("\n"+(stringindent (3))+ "verticesdata = {\n"+(stringindent (4))) to:outfile
					-- export verts
					for v = 1 to trimesh.numverts do( 
						local outpos = (in coordsys local getVert trimesh v)
						outpos *= objscale				
						format "%" ("{" + (outpos.x as string)+ "," + (outpos.y as string) + "," + (outpos.z as string)+ "}") to:outfile 
						if ( v < trimesh.numverts ) then(
							format "%" (",") to:outfile
						)
						if ((mod v 6) == 0) then(
							format "%" ("\n"+(stringindent (4))) to:outfile
						)
					)
					format "%" ("},\n"+(stringindent (3))+ "indicesdata = {\n"+(stringindent (4))) to:outfile
					-- export faces
					for f = 1 to trimesh.numfaces do (
						local index = getFace trimesh f
						index -= 1
						format "%" ((index.x as string)+ "," + (index.y as string) + "," + (index.z as string)) to:outfile 
						if ( f < trimesh.numfaces ) then(
							format "%" (",") to:outfile
						)
						if ((mod f 6) == 0) then(
							format "%" ("\n"+(stringindent (4))) to:outfile
						)
					)
					
					format "%" ("}\n"+(stringindent (2))) to:outfile
				)
			)
			
			format "%" ("}},\n") to:outfile
			
			col = custAttributes.get obj colnode baseObject:true
			if (col != undefined ) then
			(
				-- custom	
			)
		)
		
		format "%" ("}\n") to:outfile
		format "COL Export done: %\n" outfileName
		close outfile
		)
	)

/*	group "Collision Custom Info"
	(
		button 	btAddNode	"Add" across:2
		button 	btRmNode	"Remove"
	)
	*/
	group "Collision Export"
	(
		spinner spPrescale "Global Scale" range:[0,10000,1] scale:0.001
		button	btExport	"Export Selected"
	)
	group "About"
	(
		label lb1 "Luxinia Collision Exporter" align:#center
		label lb2 "v1.0 by Christoph Kubisch" align:#center
		hyperLink lb3 "www.luxinia.de" address:"http://www.luxinia.de" align:#center
	)
	/*
	on btAddNode pressed do
	(
		for obj in selection do
		(
			class = classOf obj.baseObject
			if (class == Box or class == Sphere or class ==	Capsule or class ==	Editable_mesh or class == Editable_Poly) then
			(
				custAttributes.add obj colnode
				format "COL added: %" obj.name
			)
		)
	)
	on btRmNode pressed do
	(
		for obj in selection do
		(
			col = custAttributes.get obj colnode baseObject:true
			if (col != undefined) then (
				custAttributes.delete obj colnode baseobject:true
				format "COL removed: %" obj.name
			)
		)
	)
	*/
	on btExport pressed do
	(
		local filename = getSaveFileName types:"Lua (*.lua)|*lua|All|*.*"
		if (filename != undefined) then (
			colExportSelected filename spPrescale.value
		)
	)

	--createdialog COLexporterRoll 190 350 style:#(#style_titlebar, #style_sysmenu, #style_resizing, #style_minimizebox)
)